// code for REML fitting an OUSS process to a time series or a periodogram
#ifndef FIT_OUT
#define FIT_OU


#include <complex>
#include "Config.h"
#include "InternalDefs.h"
#include "Auxiliary.cpp"
#include "../ALGLIB/src/dataanalysis.h"
#include "../ALGLIB/src/interpolation.h"
#include "../ALGLIB/src/stdafx.h"
#include "../ALGLIB/src/linalg.h"
using namespace std;


//global variables for data-dependent objective functions
vector<double> fit_RELL_OUSS_aux_w, fit_RELL_OUSS_aux_t, fit_RELL_OUSS_aux_scratch;
alglib::real_2d_array fit_RELL_OUSS_aux_Z;
vector<double> fit_PSLL_OUSS_aux_f, fit_PSLL_OUSS_aux_p;
double fit_PS_OUSS_aux_dt, fit_PS_OUSS_aux_series_size, fit_PS_OUSS_aux_T;



#pragma mark Least-squares for periodogram
#pragma mark  

double fitted_RSS_OUSS(	const alglib::real_2d_array	&x,
						const alglib::real_1d_array	&y,
						const alglib::real_1d_array &c,
						void (*func)(const alglib::real_1d_array &c, const alglib::real_1d_array &x, double &func, void *ptr)){
	double RSS=0, ty;
	void *ptr;
	long n;
	alglib::real_1d_array tx; tx.setlength(1);
	for(n=0; n<y.length(); ++n){
		tx[0] = x[n][0];
		func(c,tx,ty,ptr);
		RSS += SQ(ty - y[n]);
	}
	return RSS;
}


inline double fit_PS_OUSS_aux(const alglib::real_1d_array &c, double x){
	// s_o=SQ(c[0]), lambda=c[1], s_e=SQ(c[2])
	return SQ(c[0]) * SQ(c[1])/(SQ(2*PI*x) + SQ(c[1])) + SQ(c[2]);
}

inline void fit_PS_OUSS_aux(const alglib::real_1d_array &c, const alglib::real_1d_array &x, double &func, void *ptr){
	func = 	fit_PS_OUSS_aux(c, x[0]);
}


void fit_grad_PS_OUSS_aux(const alglib::real_1d_array &c, const alglib::real_1d_array &x, double &func, alglib::real_1d_array &grad, void *ptr){
    // this callback calculates PS(c,x) and gradient = nabla_c PS(c,x)
	// s_o=SQ(c[0]), lambda=c[1], s_e=SQ(c[2])
    func = fit_PS_OUSS_aux(c,x[0]);
    grad[0] = 2*c[0] * SQ(c[1])/(SQ(2*PI*x[0]) + SQ(c[1]));
    grad[1] = 2 * SQ(c[0]) *c[1]*SQ(2*PI*x[0])/SQ(SQ(2*PI*x[0]) + SQ(c[1]));
    grad[2] = 2*c[1];

}

inline void fit_log_PS_OUSS_aux(const alglib::real_1d_array &c, const alglib::real_1d_array &x, double &func, void *ptr){
	func = 	log(fit_PS_OUSS_aux(c, x[0]));
}


void fit_grad_log_PS_OUSS_aux(const alglib::real_1d_array &c, const alglib::real_1d_array &x, double &func, alglib::real_1d_array &grad, void *ptr){
	fit_grad_PS_OUSS_aux(c,x,func,grad,NULL);
	grad[0] /= func;
	grad[1] /= func;
	grad[2] /= func;
	func 	 = log(func);
}




#pragma mark -
#pragma mark Least-squares for periodogram (corrected for discrete time series)
#pragma mark 



//theoretical power spectrum for OUSS process, corrected for discrete time series
inline double PS_DTS_OUSS(double lambda, double power_o, double power_e, double dt, double freq){
	const double q = exp(-lambda*dt);
	return power_o * SQ(1-q)/(1+SQ(q)-2*q*cos(2*PI*freq*dt)) + power_e;
}


inline double fit_PS_DTS_OUSS_aux(const alglib::real_1d_array &c, double x){
	// s_o=SQ(c[0]), lambda=c[1], s_e=SQ(c[2]), x=freq
	return PS_DTS_OUSS(abs(c[1]),SQ(c[0]),SQ(c[2]),fit_PS_OUSS_aux_dt,x);
}


inline void fit_PS_DTS_OUSS_aux(const alglib::real_1d_array &c, const alglib::real_1d_array &x, double &func, void *ptr){
	func = 	fit_PS_DTS_OUSS_aux(c, x[0]);
}


void fit_grad_PS_DTS_OUSS_aux(const alglib::real_1d_array &c, const alglib::real_1d_array &x, double &func, alglib::real_1d_array &grad, void *ptr){
    // this callback calculates PS(c,x) and gradient = nabla_c PS(c,x)
	// s_o=SQ(c[0]), lambda=abs(c[1]), s_e=SQ(c[2])
    func = fit_PS_DTS_OUSS_aux(c,x[0]);
    const double &dt = fit_PS_OUSS_aux_dt;
	const double q = exp(-abs(c[1])*dt);
    grad[0] = 2*c[0] 	* SQ(1-q)/(1+SQ(q)-2*q*cos(2*PI*x[0]*dt));
    grad[1] = (c[1]>0 ? 1 : (c[1]<0 ? -1 : 0))	* 2*SQ(c[0])*q*fit_PS_OUSS_aux_dt*(1-q)*(SQ(q)+q+cos(2*PI*x[0]*dt)*(1-3*q))/SQ(1+SQ(q)-2*q*cos(2*PI*x[0]*dt));
    grad[2] = 2*c[2]	* 1;

}

inline void fit_log_PS_DTS_OUSS_aux(const alglib::real_1d_array &c, const alglib::real_1d_array &x, double &func, void *ptr){
	func = 	log(fit_PS_DTS_OUSS_aux(c, x[0]));
}


void fit_grad_log_PS_DTS_OUSS_aux(const alglib::real_1d_array &c, const alglib::real_1d_array &x, double &func, alglib::real_1d_array &grad, void *ptr){
	fit_grad_PS_DTS_OUSS_aux(c,x,func,grad,NULL);
	grad[0] /= func;
	grad[1] /= func;
	grad[2] /= func;
	func 	 = log(func);
}





#pragma mark -
#pragma mark Least-squares for periodogram (corrected for discrete & finite time series)
#pragma mark  


inline double PS_DFTS_OUSS(double lambda, double power_o, double power_e, double dt, double series_size, double freq){
	// s_o=SQ(c[0]), lambda=c[1], s_e=SQ(c[2]), x=freq
	const double rho = exp(-lambda*dt);
	const double N = series_size;
	const double T = dt*(N-1);
	const complex<double> r = rho*exp(complex<double>(0,2.0*PI*freq*dt));
	return power_e + power_o*T*(1.0-rho)/(SQ(N)*dt*(1.0+rho)) * (N*(1.0-SQ(rho))/SQ(abs(1.0-r)) - 2*real(r*(1.0-pow(r,N))/SQ(1.0-r)));
}

inline double fit_PS_DFTS_OUSS_aux(const alglib::real_1d_array &c, double x){
	// s_o=SQ(c[0]), lambda=abs(c[1]), s_e=SQ(c[2]), x=freq
	return PS_DFTS_OUSS(abs(c[1]), SQ(c[0]), SQ(c[2]), fit_PS_OUSS_aux_dt, fit_PS_OUSS_aux_series_size, x);
}

inline void fit_PS_DFTS_OUSS_aux(const alglib::real_1d_array &c, const alglib::real_1d_array &x, double &func, void *ptr){
	func = fit_PS_DFTS_OUSS_aux(c, x[0]);
}


void fit_grad_PS_DFTS_OUSS_aux(const alglib::real_1d_array &c, const alglib::real_1d_array &x, double &func, alglib::real_1d_array &grad, void *ptr){
    // this callback calculates PS(c,x) and gradient = nabla_c PS(c,x)
	// s_o=SQ(c[0]), lambda=abs(c[1]), s_e=SQ(c[2])
    func = fit_PS_DFTS_OUSS_aux(c,x[0]);
    const double dt 	= fit_PS_OUSS_aux_dt;
	const double N 		= fit_PS_OUSS_aux_series_size;
	const double rho 	= exp(-abs(c[1])*dt);
	const complex<double> A = exp(complex<double>(0,+2*PI*x[0]*dt));
	const complex<double> B = exp(complex<double>(0,-2*PI*x[0]*dt));
	const complex<double> r = rho*A;
	const double power_o = SQ(c[0]);
    grad[0] = 2*c[0] 	* fit_PS_OUSS_aux_T*(1.0-rho)/(SQ(N)*dt*(1.0+rho)) * (N*(1.0-SQ(abs(r)))/SQ(abs(1.0-r)) - 2*real(r*(1.0-pow(r,N))/SQ(1.0-r)));
    grad[1] = (c[1]>0 ? 1 : (c[1]<0 ? -1 : 0))	* rho * (0-dt) * real(	(power_o*fit_PS_OUSS_aux_T*((2.0*N + A*((-1.0 + pow(A*rho,N))*(-1.0 + rho*(2.0 + rho + A*(-1.0 + (rho-2.0)*rho))) + 
																		N*(-1.0 - pow(A*rho,N) + rho*(-4.0 + A + rho + 2.0*A*rho - A*SQ(rho) + 
																		pow(A*rho,N)*(A + rho - A*SQ(rho))))))/Qube(-1.0 + A*rho) + 
																		(B*((-1.0 + N)*(-1.0 + rho*(2.0 + rho)) + 2*N*Qube(rho)*SQ(B) + 
																		(-1.0 + rho*(2.0 + rho) + N*(-1.0 + SQ(rho)))*pow(rho*B,N) + 
																		rho*B*((-1.0 + (-2 + rho)*rho)*(-1.0 + pow(rho*B,N)) + 
																		N*(1.0 + pow(rho*B,N) - rho*(4.0 + rho + rho*pow(rho*B,N))))))/
																		Qube(-1.0 + rho*B)))/(dt*SQ(N)*SQ(1.0 + rho)));
    grad[2] = 2*c[2]	* 1;
}


inline void fit_log_PS_DFTS_OUSS_aux(const alglib::real_1d_array &c, const alglib::real_1d_array &x, double &func, void *ptr){
	func = 	log(fit_PS_DFTS_OUSS_aux(c, x[0]));
}


void fit_grad_log_PS_DFTS_OUSS_aux(const alglib::real_1d_array &c, const alglib::real_1d_array &x, double &func, alglib::real_1d_array &grad, void *ptr){
	fit_grad_PS_DFTS_OUSS_aux(c,x,func,grad,NULL);
	grad[0] /= func;
	grad[1] /= func;
	grad[2] /= func;
	func 	 = log(func);
}




#pragma mark -
#pragma mark Log-Likelihood for periodogram (PSLL)
#pragma mark


//negated log-likelihood for OUSS periodogram
double fit_PSLL_OUSS_aux(const alglib::real_1d_array &x){
	double LL = 0;
	double PSexpected;
	long M = fit_PSLL_OUSS_aux_f.size();
	for(long i=0; i<M; ++i){
		PSexpected = fit_PS_OUSS_aux(x, fit_PSLL_OUSS_aux_f[i]);
		LL	+= log(PSexpected) + fit_PSLL_OUSS_aux_p[i]/PSexpected;
	}
//	cout << "  LL " << LL << endl;
	return LL;
}


void fit_PSLL_OUSS_aux(const alglib::real_1d_array &x, double &func, void *ptr){
	func = fit_PSLL_OUSS_aux(x);
}



// this callback calculates PSLL(c,x) and gradient = nabla_c PSLL(c,x)
void fit_grad_PSLL_OUSS_aux(const alglib::real_1d_array &x, double &func, alglib::real_1d_array &grad, void *ptr){
    long M = fit_PSLL_OUSS_aux_f.size();
    double factor, PSexpected;
    alglib::real_1d_array freq = "[0]";
    alglib::real_1d_array tgrad; tgrad.setlength(3);
    grad[0] = grad[1] = grad[2] = func = 0;
    for(long i=0; i<M; ++i){
    	freq[0]	= fit_PSLL_OUSS_aux_f[i];
    	fit_grad_PS_OUSS_aux(x, freq, PSexpected, tgrad, ptr);
    	factor 	= 1/PSexpected - fit_PSLL_OUSS_aux_p[i]/SQ(PSexpected);
		grad[0] += factor * tgrad[0];
		grad[1] += factor * tgrad[1];
		grad[2] += factor * tgrad[2];
		func	+= log(PSexpected) + fit_PSLL_OUSS_aux_p[i]/PSexpected;
	}
}




#pragma mark -
#pragma mark Log-Likelihood for periodogram (PSLL) (corrected for discrete time series)
#pragma mark


//negated log-likelihood for OUSS periodogram
double fit_PSLL_DTS_OUSS_aux(const alglib::real_1d_array &x){
	double LL = 0;
	double PSexpected;
	long M = fit_PSLL_OUSS_aux_f.size();
	for(long i=0; i<M; ++i){
		PSexpected = fit_PS_DTS_OUSS_aux(x, fit_PSLL_OUSS_aux_f[i]);
		LL	+= log(PSexpected) + fit_PSLL_OUSS_aux_p[i]/PSexpected;
	}
//	cout << "  LL " << LL << endl;
	return LL;
}


void fit_PSLL_DTS_OUSS_aux(const alglib::real_1d_array &x, double &func, void *ptr){
	func = fit_PSLL_DTS_OUSS_aux(x);
}



// this callback calculates PSLL(c,x) and gradient = nabla_c PSLL(c,x)
void fit_grad_PSLL_DTS_OUSS_aux(const alglib::real_1d_array &x, double &func, alglib::real_1d_array &grad, void *ptr){
    long M = fit_PSLL_OUSS_aux_f.size();
    double factor, PSexpected;
    alglib::real_1d_array freq = "[0]";
    alglib::real_1d_array tgrad; tgrad.setlength(3);
    grad[0] = grad[1] = grad[2] = func = 0;
    for(long i=0; i<M; ++i){
    	freq[0]	= fit_PSLL_OUSS_aux_f[i];
    	fit_grad_PS_DTS_OUSS_aux(x, freq, PSexpected, tgrad, ptr);
    	factor 	= 1/PSexpected - fit_PSLL_OUSS_aux_p[i]/SQ(PSexpected);
		grad[0] += factor * tgrad[0];
		grad[1] += factor * tgrad[1];
		grad[2] += factor * tgrad[2];
		func	+= log(PSexpected) + fit_PSLL_OUSS_aux_p[i]/PSexpected;
	}
}





#pragma mark -
#pragma mark Log-Likelihood for periodogram (PSLL) (corrected for discrete & finite time series)
#pragma mark


//negated log-likelihood for OUSS periodogram
double fit_PSLL_DFTS_OUSS_aux(const alglib::real_1d_array &x){
	double LL = 0;
	double PSexpected;
	long M = fit_PSLL_OUSS_aux_f.size();
	for(long i=0; i<M; ++i){
		PSexpected = fit_PS_DFTS_OUSS_aux(x, fit_PSLL_OUSS_aux_f[i]);
		LL	+= log(PSexpected) + fit_PSLL_OUSS_aux_p[i]/PSexpected;
	}
//	cout << "  LL " << LL << endl;
	return LL;
}


void fit_PSLL_DFTS_OUSS_aux(const alglib::real_1d_array &x, double &func, void *ptr){
	func = fit_PSLL_DFTS_OUSS_aux(x);
}



// this callback calculates PSLL(c,x) and gradient = nabla_c PSLL(c,x)
void fit_grad_PSLL_DFTS_OUSS_aux(const alglib::real_1d_array &x, double &func, alglib::real_1d_array &grad, void *ptr){
    long M = fit_PSLL_OUSS_aux_f.size();
    double factor, PSexpected;
    alglib::real_1d_array freq = "[0]";
    alglib::real_1d_array tgrad; tgrad.setlength(3);
    grad[0] = grad[1] = grad[2] = func = 0;
    for(long i=0; i<M; ++i){
    	freq[0]	= fit_PSLL_OUSS_aux_f[i];
    	fit_grad_PS_DFTS_OUSS_aux(x, freq, PSexpected, tgrad, ptr);
    	factor 	= 1/PSexpected - fit_PSLL_OUSS_aux_p[i]/SQ(PSexpected);
		grad[0] += factor * tgrad[0];
		grad[1] += factor * tgrad[1];
		grad[2] += factor * tgrad[2];
		func	+= log(PSexpected) + fit_PSLL_OUSS_aux_p[i]/PSexpected;
	}
}


#pragma mark -
#pragma mark Log-Likelihood for time series
#pragma mark


// calculate D*V*D^T, where V is quadratic of size M
// D is upper 2-band matrix, D=[[-1,1,0,0,..],[0,-1,1,0,..],[0,0,-1,1,..],...]
// scratch[] is of size (M-1)*M (must be pre-allocated, used only for intermediate storage)
// result is returned in V[] and is a quadratic matrix of size (M-1)*(M-1)
void fit_RELL_OUSS_matmul_DVD_aux(alglib::real_2d_array& V, long M, double scratch[]){
	long r,c,k;
	
	//D*V
	for(r=0; r<(M-1); ++r){
		for(c=0; c<M; ++c){
			scratch[r*M+c] = V[r+1][c] - V[r][c];
		}
	}
	
	//(D*V)*D^T
	for(r=0; r<(M-1); ++r){
		for(c=0; c<(M-1); ++c){
			V[r][c] = scratch[r*M+(c+1)] - scratch[r*M+c];
		}
	}
}


//calculates w^T * A * w
double fit_RELL_OUSS_scalarproduct(const alglib::real_2d_array& A, long n, const vector<double> &w){
	vector<double> y(n);
	long r,c;
	for(r=0; r<n; ++r){
		for(y[r]=0, c=0; c<n; ++c){
			y[r] += A[r][c]*w[c];
		}
	}
	double sum=0;
	for(r=0; r<n; ++r){
		sum += w[r] * y[r];
	}
	return sum;
}



void print_2d_array(const alglib::real_2d_array& A, long M){
	cout << "Array " << M << "x" << M << ":\n";
	long r,c;
	for(r=0; r<M; ++r){
		for(c=0; c<M; ++c){
			cout << A[r][c] << "  ";
		}
		cout << "\n";
	}
	cout << "\n";
}


//calculate log-determinant of a real symmetric matrix
double fit_RELL_OUSS_aux_logdet(const alglib::real_2d_array& A, long M){
	alglib::real_1d_array lambdas;
	alglib::real_2d_array dummy;
	alglib::smatrixevd(A,M,0,true,lambdas,dummy);
	double logdet=0;
	for(long n=0; n<M; ++n){ logdet += log(lambdas[n]); }
	return logdet;
}



//returns negated and restricted log-likelihood for OUSS with parameters x[]
//see [Dennis et al (2014) - Density dependent state-space model for population abundance data with unequal time intervals, Eq. 22]
double fit_RELL_OUSS_aux(const alglib::real_1d_array &x){
	double lambda	= abs(x[0]);
	double sigma 	= x[1];
	double epsilon 	= x[2];
	long M = fit_RELL_OUSS_aux_w.size()+1;
	alglib::real_2d_array &Z=fit_RELL_OUSS_aux_Z;
	long r,c;
		
	//define V
	for(r=0; r<M; ++r){
		for(c=0; c<M; ++c){
			Z[r][c] = (r==c ? SQ(epsilon) : 0) + SQ(sigma)/(2*lambda) * exp(-lambda*abs(fit_RELL_OUSS_aux_t[r]-fit_RELL_OUSS_aux_t[c]));
		}
	}

	//replace V with D*Z*D^T
	fit_RELL_OUSS_matmul_DVD_aux(Z, M, &fit_RELL_OUSS_aux_scratch[0]);

	//log-determinant of Z
	double Zlogdet = fit_RELL_OUSS_aux_logdet(Z, M-1);
	
	//inverse Z --> replace old Z
	alglib::ae_int_t info;
	alglib::matinvreport report;
	alglib::spdmatrixinverse(Z, M-1, true, info, report);
	if(info<0){ cout << "  ERROR during Zinv in fit_RELL_OUSS_aux(): Z nearly singular\n"; } //debug
	
	//negated restricted log-likelihood
	double LL = (M-1.0)/2 * log(2*PI) + 0.5*Zlogdet + 0.5*fit_RELL_OUSS_scalarproduct(Z, M-1, fit_RELL_OUSS_aux_w);
	
	return LL;
}



void fit_RELL_OUSS_aux(const alglib::real_1d_array &x, double &func, void *ptr){
	func = fit_RELL_OUSS_aux(x);
}


#pragma mark -
#pragma mark Fitting routines
#pragma mark 


//least-squares fit of OUSS periodogram
bool fit_PS_LS_OUSS(double 					df,
					double					dt,	// original time series time step
					long					N, 	// original time series size
					const vector<double> 	&spectrum,
					double					suspected_lambda,
					double					suspected_power_o,
					double					suspected_power_e,
					double					&fitted_lambda,
					double					&fitted_power_o,
					double					&fitted_power_e){
	long count = spectrum.size()-1;
	long n,m,k,q,p,r;
	alglib::real_2d_array x;
	x.setlength(count, 1);
	for(n=1; n<spectrum.size(); ++n){ x[n-1][0] = n*df; }
	fit_PS_OUSS_aux_dt = dt;
	fit_PS_OUSS_aux_T = dt*(N-1.0);
	fit_PS_OUSS_aux_series_size = N;
	
    alglib::real_1d_array y;
    y.setlength(count);
    for(n=1; n<spectrum.size(); ++n){ y[n-1] = (PS_OU_FIT_LOGARITHMICALLY ? log(spectrum[n]) : spectrum[n]); }
    
    alglib::real_1d_array c,oc;
    oc.setlength(3);
	oc[0] = sqrt(abs(suspected_power_o));	//power_o
	oc[1] = abs(suspected_lambda);			//lambda
	oc[2] = sqrt(abs(suspected_power_e));	//power_e
	
	alglib::real_1d_array scales;
	scales.setlength(oc.length());
	for(n=0; n<oc.length(); ++n) scales[n] = oc[n];

	
    double epsf = PS_FIT_RELATIVE_ACCURACY*(PS_OU_FIT_LOGARITHMICALLY ? log(maximum(spectrum,1,spectrum.size()-1))-log(minimum(spectrum,1,spectrum.size()-1)) : maximum(spectrum,1,spectrum.size()-1));
    double epsx = df/10000;
    alglib::ae_int_t maxits = PS_OU_ALGLIB_FIT_MAX_ITERATIONS;
    alglib::ae_int_t info;
    alglib::lsfitstate state;
    alglib::lsfitreport rep;
    double diffstep = min(df/10000, suspected_lambda/10000);
    
    vector<double> trials_lambda, trials_power_o, trials_power_e;
    for(n=-PS_OU_FIT_PARAMS_TRY_PLUSMINUS_ORDERDOFMAGNITUDE; n<=PS_OU_FIT_PARAMS_TRY_PLUSMINUS_ORDERDOFMAGNITUDE; ++n){
    	trials_power_o.push_back(oc[0]*pow(3, 1.0*n));
    }
    for(n=-PS_OU_FIT_PARAMS_TRY_PLUSMINUS_ORDERDOFMAGNITUDE; n<=PS_OU_FIT_PARAMS_TRY_PLUSMINUS_ORDERDOFMAGNITUDE; ++n){
    	trials_lambda.push_back(oc[1]*pow(3, 1.0*n));
    }
	trials_power_e.push_back(oc[2]); // power_e is a uniform translation parameter and so very easy to deal with by the optimizer 

    vector<double> trials_RSS;
    vector<alglib::real_1d_array> trials_c;
    
	for(p=0; p<trials_power_o.size(); ++p){
		for(q=0; q<trials_lambda.size(); ++q){
			for(r=0; r<trials_power_e.size(); ++r){
				c = oc;
				c[0] = trials_power_o[p];
				c[1] = trials_lambda[q];
				c[2] = trials_power_e[r];
				
				try{
					if(FIT_PS_USE_GRADIENT){ alglib::lsfitcreatefg(x, y, c, true, state); //define data and function family
					}else{ alglib::lsfitcreatef(x, y, c, diffstep, state); }
					alglib::lsfitsetcond(state, epsf, epsx, maxits); //stoping conditions
					alglib::lsfitsetscale(state, scales); //define parameter scales								
					//fit:
					if(CORRECT_FOR_DISCRETE_FINITE_TIME_SERIES){
						if(FIT_PS_USE_GRADIENT){
							if(PS_OU_FIT_LOGARITHMICALLY){ alglib::lsfitfit(state, fit_log_PS_DFTS_OUSS_aux, fit_grad_log_PS_DFTS_OUSS_aux); }
							else{ alglib::lsfitfit(state, fit_PS_DFTS_OUSS_aux, fit_grad_PS_DFTS_OUSS_aux); }
						}else{ 
							if(PS_OU_FIT_LOGARITHMICALLY){ alglib::lsfitfit(state, fit_log_PS_DFTS_OUSS_aux); }
							else{ alglib::lsfitfit(state, fit_PS_DFTS_OUSS_aux); }
						}
					}else{
						if(FIT_PS_USE_GRADIENT){
							if(PS_OU_FIT_LOGARITHMICALLY){ alglib::lsfitfit(state, fit_log_PS_OUSS_aux, fit_grad_log_PS_OUSS_aux); }
							else{ alglib::lsfitfit(state, fit_PS_OUSS_aux, fit_grad_PS_OUSS_aux); }
						}else{ 
							if(PS_OU_FIT_LOGARITHMICALLY){ alglib::lsfitfit(state, fit_log_PS_OUSS_aux); }
							else{ alglib::lsfitfit(state, fit_PS_OUSS_aux); }
						}
					}
					alglib::lsfitresults(state, info, c, rep);   //read fit results						
				}
				catch (...) { return false; }
				if(int(info)>0){
					trials_c.push_back(c);
					if(CORRECT_FOR_DISCRETE_FINITE_TIME_SERIES){
						if(PS_OU_FIT_LOGARITHMICALLY){	trials_RSS.push_back(fitted_RSS_OUSS(x, y, c, fit_log_PS_DFTS_OUSS_aux)); }
						else{ trials_RSS.push_back(fitted_RSS_OUSS(x, y, c, fit_PS_DFTS_OUSS_aux)); }
					}else{
						if(PS_OU_FIT_LOGARITHMICALLY){	trials_RSS.push_back(fitted_RSS_OUSS(x, y, c, fit_log_PS_OUSS_aux)); }
						else{ trials_RSS.push_back(fitted_RSS_OUSS(x, y, c, fit_PS_OUSS_aux)); }
					}
				}
			}
		}
	}
	
	//return best fit
	if(trials_RSS.empty()) return false;
	long bestTrial	= minimumIndex(trials_RSS);
	c 				= trials_c[bestTrial];
    fitted_power_o 	= SQ(c[0]);
    fitted_lambda 	= abs(c[1]);
    fitted_power_e 	= SQ(c[2]);
    return true;
}








//maximum-likelihood fit of OUSS periodogram
bool fit_PS_ML_OUSS(double 					df,
					double 					dt,
					long					N, 	// original time series size
					const vector<double> 	&spectrum,
					double					suspected_lambda,
					double					suspected_power_o,
					double					suspected_power_e,
					double					&fitted_lambda,
					double					&fitted_power_o,
					double					&fitted_power_e){
	long count = spectrum.size()-1;
	long n,m,k,q,p,r;
	
	//store data for ML fit
	fit_PSLL_OUSS_aux_f.resize(count); 
	fit_PSLL_OUSS_aux_p.resize(count);
	for(n=1; n<spectrum.size(); ++n){ fit_PSLL_OUSS_aux_f[n-1] = n*df; fit_PSLL_OUSS_aux_p[n-1] = spectrum[n]; }
	fit_PS_OUSS_aux_dt = dt;
	fit_PS_OUSS_aux_T = dt*(N-1.0);
	fit_PS_OUSS_aux_series_size = N;
	
	//scales
    alglib::real_1d_array scales;
    scales.setlength(3);
	scales[0] = sqrt(abs(suspected_power_o));	//power_o
	scales[1] = abs(suspected_lambda);			//lambda
	scales[2] = sqrt(abs(suspected_power_e));	//power_e
		
	//precision and stopping conditions
	double epsg = 1e-10; 	// min gradient (dx is taken relative to scales)
	double epsf = 0; 		// min function step
	double epsx = 1e-3; 	// spatial step (relative to scales)
	double dx	= 1e-3; 	// differentiation step (relative to scales) if gradient is not provided
	alglib::ae_int_t maxits = PS_OU_ALGLIB_FIT_MAX_ITERATIONS;
	
	//allocate
	alglib::minlbfgsstate state;
	alglib::minlbfgsreport report;
	alglib::real_1d_array xfinal, xinit;
	xfinal = xinit = scales;
	vector<double> trials_ML;
    vector<alglib::real_1d_array> trials_xfinal;
    alglib::real_1d_array LL = "[0]";
	
	//expand grid of initial guesses
	vector<vector<double> > trials(scales.length());
    for(n=-PS_OU_FIT_PARAMS_TRY_PLUSMINUS_ORDERDOFMAGNITUDE; n<=PS_OU_FIT_PARAMS_TRY_PLUSMINUS_ORDERDOFMAGNITUDE; ++n){
    	for(m=0; m<trials.size(); ++m){
			trials[m].push_back(xinit[m]*pow(3.0, 1.0*n));
    	}
    }
    trials[2].clear(); trials[2].push_back(xinit[2]); // power_e is a uniform translation parameter and so very easy to deal with by the optimizer 
	
	//search for minimum starting at different guesses
	for(q=0; q<trials[0].size(); ++q){
		for(p=0; p<trials[1].size(); ++p){
			for(r=0; r<trials[2].size(); ++r){
				xinit[0] = trials[0][q];
				xinit[1] = trials[1][p];
				xinit[2] = trials[2][r];
				try{
					if(FIT_PS_USE_GRADIENT){ alglib::minlbfgscreate(3, xinit, state);
					}else{ alglib::minlbfgscreatef(3, xinit, dx, state); }
					alglib::minlbfgssetcond(state, epsg, epsf, epsx, maxits);
					alglib::minlbfgssetscale(state, scales);
					
					if(CORRECT_FOR_DISCRETE_FINITE_TIME_SERIES){
						if(FIT_PS_USE_GRADIENT){ alglib::minlbfgsoptimize(state, fit_grad_PSLL_DFTS_OUSS_aux);
						}else{ alglib::minlbfgsoptimize(state, fit_PSLL_DFTS_OUSS_aux); }
					}else{
						if(FIT_PS_USE_GRADIENT){ alglib::minlbfgsoptimize(state, fit_grad_PSLL_OUSS_aux);
						}else{ alglib::minlbfgsoptimize(state, fit_PSLL_OUSS_aux); }
					}
					alglib::minlbfgsresults(state, xfinal, report);
				}
				catch (...) { return false; }
				if(int(report.terminationtype)>0){
					//keep record of fit results
					trials_xfinal.push_back(xfinal);
					trials_ML.push_back(CORRECT_FOR_DISCRETE_FINITE_TIME_SERIES ? fit_PSLL_DFTS_OUSS_aux(xfinal) : fit_PSLL_OUSS_aux(xfinal));
//					cout << "  ...  LL=" << trials_ML.back() << endl;
				}
			}
		}
	}
	
	//return best fit
	if(trials_ML.empty()) return false;
	long bestTrial	= minimumIndex(trials_ML);
	xfinal 			= trials_xfinal[bestTrial];
    fitted_power_o 	= SQ(xfinal[0]);
    fitted_lambda 	= abs(xfinal[1]);
    fitted_power_e 	= SQ(xfinal[2]);
    return true;
}











// Restricted Maximum Likelihood fit OUSS process to a time series
// dX = -theta*(X-mu)*dt + sigma*dW, where W is the standard Wiener process
// Y = X + epsilon*N, where N is standard-normally distributed
// See [Dennis et al (2014) - Density dependent state-space model for population abundance data with unequal time intervals]
bool fit_TS_ML_OUSS(const vector<double>	&times,
					const vector<double> 	&values,
					double					suspected_lambda,
					double					suspected_sigma,
					double					suspected_epsilon, // measurement error for state-space model
					double					&fitted_mu,
					double					&fitted_lambda,
					double					&fitted_sigma,
					double					&fitted_epsilon){
	
	long n,i,q,p,r;
	const long M = values.size();
	
	//preprocess data for ML
	fit_RELL_OUSS_aux_w.resize(M-1);
	for(n=1; n<M; ++n){
		fit_RELL_OUSS_aux_w[n-1] = values[n] - values[n-1];
	}
	fit_RELL_OUSS_aux_t = times;
	fit_RELL_OUSS_aux_scratch.resize((M-1)*M);
	fit_RELL_OUSS_aux_Z.setlength(M,M);
	
	//scales
	alglib::real_1d_array scales;
	scales.setlength(3);
	scales[0] = suspected_lambda;
	scales[1] = suspected_sigma;
	scales[2] = suspected_epsilon;
		
	//precision and stopping conditions
	double epsg = 1e-8; // min gradient (dx is taken relative to scales)
	double epsf = 0; 	//min function step
	double epsx = 1e-2; // spatial step (relative to scales)
	double dx	= 1e-3; // differentiation step (relative to scales)
	alglib::ae_int_t maxits = PS_OU_ALGLIB_FIT_MAX_ITERATIONS;
	
	//allocate
	alglib::minlbfgsstate state;
	alglib::minlbfgsreport report;
	alglib::real_1d_array xfinal, xinit;
	xfinal = xinit = scales;
	vector<double> trials_ML;
    vector<alglib::real_1d_array> trials_xfinal;
    alglib::real_1d_array LL = "[0]";
	
	//expand grid of initial guesses
	vector<double> trials_lambda, trials_sigma, trials_epsilon;
    for(n=-PS_OU_FIT_PARAMS_TRY_PLUSMINUS_ORDERDOFMAGNITUDE; n<=PS_OU_FIT_PARAMS_TRY_PLUSMINUS_ORDERDOFMAGNITUDE; ++n){
    	trials_lambda.push_back(xinit[0]*pow(3.0, 1.0*n));
    	trials_sigma.push_back(xinit[1]*pow(3.0, 1.0*n));
    	trials_epsilon.push_back(xinit[2]*pow(3.0, 1.0*n));
    }
	
	//search for minimum starting at different guesses
	for(q=0; q<trials_lambda.size(); ++q){
		for(p=0; p<trials_sigma.size(); ++p){
			for(r=0; r<trials_epsilon.size(); ++r){
				xinit[0] = trials_lambda[q];
				xinit[1] = trials_sigma[p];
				xinit[2] = trials_epsilon[r];

				try{
					//optimize log-likelihood
					alglib::minlbfgscreatef(3, xinit, dx, state);
					alglib::minlbfgssetcond(state, epsg, epsf, epsx, maxits);
					alglib::minlbfgssetscale(state, scales);
					alglib::minlbfgsoptimize(state, fit_RELL_OUSS_aux);
					alglib::minlbfgsresults(state, xfinal, report);
				}
				catch (...) { return false; }
				if(int(report.terminationtype)>0){
					//keep record of fit results
					trials_xfinal.push_back(xfinal);
					trials_ML.push_back(fit_RELL_OUSS_aux(xfinal));
				}
			}
		}
	}
	
	//return best fit
	if(trials_ML.empty()) return false;
	long bestTrial	= minimumIndex(trials_ML);
	xfinal 			= trials_xfinal[bestTrial];
    fitted_lambda 	= abs(xfinal[0]);
    fitted_sigma 	= abs(xfinal[1]);
    fitted_epsilon 	= abs(xfinal[2]);
//    cout << "    OUSS TS ML fit: lambda=" << fitted_lambda << ",  sigma=" << fitted_sigma << ",  epsilon=" << fitted_epsilon << endl; //debug
    return true;
}


#endif